from .Dictionary import Dictionary
from os import path
import re
import logging
from pprint import pprint

PATH_TO_FILE = path.join(path.dirname(__file__), 'cedict_ts.u8')

class CEDictionary(Dictionary):

    def __init__(self, debug_level):
        """Chinese to English Dictionary

        """
        self.logger = logging.getLogger('autoanki.cedict')
        self.logger.setLevel(debug_level)
        self.dict = self._parse_file()
        
        pass

    def _parse_file(self) -> dict[str, dict]:
        self.logger.info("Parsing dict file...")
        if not path.isfile(PATH_TO_FILE):
            self.logger.critical("Could not open dictionary file")

        # TODO Some entries use a dot in the midde
        #   e.g. 哈利·波特
        #   this does not currently match
        definitions = {}
        last_word = "~~"
        current_word = ""

        with open(PATH_TO_FILE, "r") as dict_file:
            for line in dict_file:
                if line[0] == '#':
                    continue
                parts = line.split(" ")
                trad_word = parts[0]
                current_word = parts[1]
                
                m = re.compile(r"\[[^]]*\]")
                match = m.findall(line)
                # self.logger.debug(match)
                pinyin_numbers = match[0]

                definition = line[line.find("]")+2:]

                if current_word != last_word:
                    # self.logger.debug(f"Creating entry: {current_word}")
                    definitions[current_word] = {'trad_word': "", 'pinyin_numbers':"", 'definition':""}
                
                # self.logger.debug(f"Word: {current_word}")
                # self.logger.debug(f"Trad: {trad_word}")
                # self.logger.debug(f"PinYin: {pinyin_numbers}")
                # self.logger.debug(f"DEF: {definition}")
                definitions[current_word]['trad_word'] = trad_word
                definitions[current_word]['pinyin_numbers'] += pinyin_numbers 
                definitions[current_word]['definition'] += definition.strip("\n") 

                last_word = current_word

        self.logger.info("Done!")
        for key, value in definitions.items():
            print(key, '->', value)
        return definitions


    def find_word(self, word: str) -> None | list[str]:

        if word in self.dict:
            """
            traditional_script = params[0]\n
            word_type = params[1]\n
            pinyin = params[2]\n
            pinyin_numbers = params[3]\n
            sub_components = params[4]\n
            hsk_level = params[5]\n
            top_level = params[6]\n
            definition = params[7]\n
            word = params[8]
            """
            params = ["", "", "", "", "", "", "", "", ""]
            params[0] = self.dict[word]['trad_word'] 
            params[3] = self.dict[word]['pinyin_numbers'] 
            params[7] = self.dict[word]['definition'] 
            params[8] = word
            return params
        else:
            return None 


    def size(self):
        with open(PATH_TO_FILE,"r") as f:
            return len(f.readlines()) 







import logging
import time
import urllib

import requests
from bs4 import BeautifulSoup
import urllib.parse
import pinyin as pin_to_num

from .Dictionary import Dictionary

logger = logging.getLogger('autoanki')
# logger.setLevel()


class YellowBridgeDictionary(Dictionary):

    def __init__(self):
        pass

    def find_word(self, word, cache_number=None):
        '''
        Helper function for complete_unfinished_dictionary_records() Takes a word (one or more characters),
        finds them on YellowBridge, and adds them to the dictionary
        :param word: The word to find on YellowBridge.
        :param cache_number: The secondary page number for a word with multiple definitions. (See comments in body of
        function)
        :return:
        '''
        print(repr(word))
        if word == None or word == '':
            print("There is no page on Yellowbridge page for null")
            return

        urlx = "http://www.yellowbridge.com/chinese/dictionary.php?word="
        url = urlx + urllib.parse.quote(word)
        if cache_number is not None:
            url += "&cache=" + cache_number

        response = requests.get(url, headers={
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36'})
        yellowbridge_soup = BeautifulSoup(response.content, "html.parser")

        body = yellowbridge_soup.find(id='tabbody')
        main_data = yellowbridge_soup.find(id='mainData')
        # Some definitions will show up on the page differently because they have multiple meanings/pronunciations
        # e.g 了 could be le or liao.
        # These pages actually link to other pages with a different 'cache' number
        # Default 了 page:
        # https://www.yellowbridge.com/chinese/dictionary.php?word=%E4%BA%86
        # Page for le:
        # https://www.yellowbridge.com/chinese/dictionary.php?word=%E4%BA%86&cache=5114
        # Page for laio:
        # https://www.yellowbridge.com/chinese/dictionary.php?word=%E4%BA%86&cache=5115
        # This means that it would have to be possible to have multiple rows in the dictionary with the same word column,
        # which breaks other parts of the code.
        # For now, we just grab the first cache number and use that one. Maybe a v2 FEATURE will be able to have multiple
        # definitions for the same character, but that is not a priority currently
        if main_data:
            yellowbridge_contains_exact_definition = True
        else:
            yellowbridge_contains_exact_definition = False
        if yellowbridge_contains_exact_definition:
            definition = simplified_script = traditional_script = pinyin = pinyin_num = part_of_speech = hsk_level = top_level = composing_words = None
            main_data = main_data.find_all('tr')
            # Collect all necessary information from the page:
            for row in main_data:
                row_info_type = row.find('td').getText()
                row_info = row.find_all('td')[1].getText()
                # print(row_info_type + ":")
                # print(row_info)
                if row_info_type == "English Definition":
                    definition = row_info
                elif row_info_type == "Simplified Script":
                    simplified_script = row_info
                elif row_info_type == "Traditional Script":
                    traditional_script = row_info.split("P")[0]
                elif row_info_type == "Pinyin":
                    pinyin = row_info
                    pinyin_num = pin_to_num.get(word, format="numerical")
                elif row_info_type == "Part of Speech":
                    part_of_speech = row_info
                elif row_info_type == "Proficiency Test Level":
                    proficiency_level = row_info
                    if "HSK=" in proficiency_level:
                        hsk_level = proficiency_level.split("HSK=")[1].split(";")[0]
                    if "TOP=" in proficiency_level:
                        top_level = proficiency_level.split("TOP=")[1].split(";")[0]

            word_decomposition_soup = yellowbridge_soup.find(id='wordDecomp')
            if len(word) > 1:
                composing_words_list = word_decomposition_soup.find_all('tr')
                composing_words = ""
                for composing_word in composing_words_list:
                    composing_words += str(composing_word.find_all('td')[0].find('a').getText()) + ";"

            # This is all of the information that has been collected
            # print([definition, simplified_script, traditional_script,pinyin,pinyin_num, part_of_speech,hsk_level,top_level,composing_words])
            params = [traditional_script, part_of_speech, pinyin, pinyin_num,
                      composing_words, hsk_level, top_level, definition, word]
            # Add the information for this word to the database

            return params
        else:
            # print("This is not a definition page. getting all sub-pages")
            matching_results = yellowbridge_soup.find(id='multiRow')
            # Grab first row. This is usually the most common definition
            row = matching_results.find_all('tr')[0]
            # print(row.find('a'))
            href = str(row.find('a')['href'])
            # print(href)
            cache_number_from_href = href.split("word=")[1].split("&")[1].replace("cache=", "")

            return self.find_word(word, cache_number_from_href)

    def size(self):
        return 0




import logging

from .BookCleaner import BookCleaner
from .DatabaseManager import DatabaseManager
from .Dictionary import CEDictionary
from .DeckManager import DeckManager

BLACK = "\u001b[30m"
RED = "\u001b[31m"
GREEN = "\u001b[32m"
YELLOW = "\u001b[33m"
BLUE = "\u001b[34m"
MAGENTA = "\u001b[35m"
CYAN = "\u001b[36m"
WHITE = "\u001b[37m"
RESET = "\u001b[0m"
logging.basicConfig(
    # filename='HISTORY.log',
    level=logging.WARNING,
    format=f'{GREEN}%(asctime)s{RESET} {RED}%(levelname)8s{RESET} {YELLOW}%(name)18s{RESET}: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)

class AutoAnki:

    def __init__(self, database_filepath='autoanki.db', debug_level=20):
        """
        Creates an instance of autoanki.
        This creates a book cleaner, database connection, and deck maker
        :param database_filepath: The filepath for the database
        :param logging_level: between 0 (DEBUG) and 50(CRITICAL)
        """
        self.logger = logging.getLogger('autoanki')
        self.logger.setLevel(debug_level)
        self.logger.debug(f"logger active")
        self.logger.info("Connecting to database...")

        self.book_cleaner = BookCleaner(debug_level)

        self.database_filepath = database_filepath
        if not DatabaseManager.is_database(database_filepath):
            self.logger.info("Creating database...")
            DatabaseManager.create_autoanki_db(database_filepath)
            self.logger.info("Done creating database.")
        self.database_manager = DatabaseManager(database_filepath, debug_level)

        self.dictionary = CEDictionary(debug_level)
        self.deck_manager = DeckManager(debug_level)

        self.logger.info("Connected!")

    def add_book(self, book_path: str, book_name: str = 'New Book'):
        """
        Add a directory full of files to the database
        :param book_path: The filepath to the directory that contains the files to add. e.g. lost_prince.txt
        :param book_name: The name of the book being added e.g. "Lost Prince"
        :return:
        """

        self.logger.debug(f"autoanki: Adding book from [{book_path}]")

        # Clean the book
        if not self.book_cleaner.clean(book_path):
            # logger.warning("autoanki: Unable to clean book [" + book_name + "].")
            return

        # Add the book to the database
        if not self.database_manager.add_book(book_path, book_name):
            self.logger.warning("Unable to add [" + book_name + "] to database.")
            return

        self.logger.info("autoanki: Added [" + book_path + "].")

    def complete_unfinished_definitions(self):
        """
        autoanki contains an internal definitions table that is scraped from the internet. As words are added to
        autoanki, their definitions must be found. This function passively finds definitions and adds them to the table
        :return: None
        """

        # TODO Make progress bar for unfinished records
        self.logger.info("Checking for records...")
        self.database_manager.cursor.execute("SELECT word FROM dictionary WHERE definition IS NULL")
        response_rows = self.database_manager.cursor.fetchall()
        #while len(response_rows) > 0:
        self.database_manager.cursor.execute("SELECT word FROM dictionary WHERE definition IS NULL")
        response_rows = self.database_manager.cursor.fetchall()
        if len(response_rows) > 0:
            self.logger.info("Adding " + str(len(response_rows)) + " rows to dictionary table")
            for row in response_rows:
                word = row[0]

                self.logger.debug(f"Finding: [{word}]...")
                params = self.dictionary.find_word(word)
                if not params:
                    self.logger.info(f"❌Could not find: [{word}]")
                else:
                    self.logger.info(f"✅Found: [{word}]")
                    self.database_manager.update_definition(params)

        else:
            self.logger.info("No new rows to complete in dictionary table")
            # time.sleep(2)

    def print_database_info(self):
        self.database_manager.print_info()

    @staticmethod
    def is_database(db_path):
        return DatabaseManager.is_database(db_path)

    @staticmethod
    def create_autoanki_db(db_path: str):
        DatabaseManager.create_autoanki_db(db_path)

    def create_deck(self, deck_name: str, filepath: str):
        """
        Creates a deck file in the directory of the main file.
        :return:
        """

        self.logger.info("Generating deck file [" + deck_name + ".apk]")
        words = self.database_manager.get_all_completed_definitions()

        deck_path = self.deck_manager.generate_deck_file(words, deck_name, filepath)
        if deck_path is None:
            self.logger.warning("Was not able to create deck file for [", deck_name, "]")
        else:
            self.logger.info("Generated deck file [" + deck_path + "]")

    @property
    def book_list(self):
        """
        Get a list of the books in the database
        :return: List of book names
        """
        return self.database_manager.books

    @book_list.setter
    def book_list(self, _):
        pass

    @property
    def unfinished_entries(self):
        return self.database_manager.unfinished_definitions()

    @unfinished_entries.setter
    def unfinished_entries(self, _):
        pass


if __name__ == '__main__':

    aa = AutoAnki()
    print(aa.book_list)



import os
import pprint
import re
import sqlite3
import logging
import unicodedata

from pathlib import Path
import jieba

FILTERED_WORDS = [
    '\n',
    ' ',
    '。',
    '"',
    "‘",
    "’",
    "“",
    "”",
    "，",
    "、",
    "·",
]

class DatabaseManager:

    def __init__(self, database_path, debug_level):
        self.logger = logging.getLogger('autoanki.dbmngr')
        self.logger.setLevel(debug_level)
        if not os.path.exists(database_path):
            self.logger.warning("The database [", database_path, "] does not exist.")
            raise Exception("Cannot create DatabaseManager with invalid database path.")
        self.database_path = database_path
        self.books = []
        path = os.path.join(os.getcwd(), self.database_path)
        self.connection = sqlite3.connect(path)
        self.cursor = self.connection.cursor()

    @staticmethod
    def convert_to_tablename(name: str):
        """
        Converts a string to a sql-valid table name.
        `param name` The name to convert
        `return` A tablename valid for an sql table
        """
        value = unicodedata.normalize('NFKC', name)
        # value.replace("：",":")
        value = value.replace("：", "__")
        value = re.sub(r'[^\w\s-]', '', value.lower())
        return re.sub(r'[-\s]+', '_', value).strip('-_')

    @staticmethod
    def is_database(database_name:str):
        if not database_name.endswith(".db"):
            return False
        if not os.path.exists(database_name):
            return False
        try:
            connection = sqlite3.connect(database_name)
            cursor = connection.cursor()
            # This will fail if dictionary table does not exist
            cursor.execute("SELECT word FROM dictionary")
            connection.close()
        except sqlite3.OperationalError:
            return False
        return True

    @staticmethod
    def create_autoanki_db(database_path):
        """
        Creates an autoanki database file, including all tables needed for autoanki
        `database_path` The path to the database to create
        """
        logger = logging.getLogger('autoanki.dbmngr')
        logger.info("Creating database [" + database_path + "]")
        path = os.path.join(os.path.dirname(__file__), 'databases_init.sql')
        try:
            with open(path, 'r') as sql_file:
                sql_script = sql_file.read()
            connection = sqlite3.connect(database_path)
            cursor = connection.cursor()
            cursor.executescript(sql_script)
            connection.commit()
        except FileNotFoundError:
            logger.warning("Could not create database: Missing SQL files")
            return False

    def _create_book_table(self, book_name, table_name) -> bool:
        """
        Creates a new entry in the book_list table
        `table_name` The name of the new entry to create
        `return` False if error
        """
        self.cursor.execute("SELECT table_name FROM book_list")
        book_list = self.cursor.fetchall()

        if (table_name,) in book_list:
            self.logger.warning("The book is already in database. Not adding")
            return False

        # self.logger.debug("Inserting")
        self.cursor.execute(f"INSERT INTO book_list VALUES(\"{book_name}\",\"{table_name}\",'cn')")
        self.connection.commit()
        fname = 'book_table.sql'
        this_file = os.path.abspath(__file__)
        this_dir = os.path.dirname(this_file)
        wanted_file = os.path.join(this_dir, fname)
        fd = open(wanted_file, 'r')
        book_table_file = fd.read()
        book_table_file = book_table_file.replace("BOOK_NAME", table_name)
        fd.close()
        # Create the new table
        self.cursor.execute(book_table_file)
        self.connection.commit()
        return True

    def add_file_to_database(self, filepath, table_name):
        """
        Adds every word in a file to both the dictionary table and the book's table
        `filepath` The path to the file
        `table_name` The name of the table to add the words to.
            This should be the same for every wile in a given book
        """
        self.logger.info(f"Adding file {filepath} to database...")

        # Get number of appearances for each word in the file, and put it into a dictionary
        word_appearances = {}
        # As we add words to the definitions table, get the id. This is for the book table
        word_ids = {}
        with open(filepath,'r',encoding='utf-8') as f:
            line = " "
            while line:
                line = f.readline()
                if line:
                    # print("Line: ", line)
                    tokenized_line = jieba.lcut(line)
                    # print("Tokenized line: ", tokenized_line)
                    for word in tokenized_line:

                        is_ascii = len(word) == len(word.encode())
                        if word not in FILTERED_WORDS and not is_ascii:
                            # print("Word: ",word)
                            if word_appearances.get(word) == None:
                                word_appearances[word] = 1
                            else:
                                word_appearances[word] += 1

        self.logger.info(f"Found {str(len(word_appearances.items()))} words in file.")

        # Add the words to the dictionary if they are not already there
        self.cursor.execute(f"SELECT word FROM dictionary")
        self.connection.commit()
        dictionary_words = self.cursor.fetchall()

        self.logger.info(f"Found {str(len(dictionary_words))} words in dictionary.")

        for word, appearances in word_appearances.items():
            if (word,) not in dictionary_words:
                # self.logger.debug("Adding word...")

                self.cursor.execute(f"INSERT INTO dictionary (word) VALUES (?)", [word])
                self.connection.commit()

        # Make a dictionary of word ids from dictionary
        self.cursor.execute(f"SELECT word_id, word FROM dictionary")
        result = self.cursor.fetchall()
        word_id_dict = {}
        for line in result:
            word_id_dict[line[1]] = line[0]

        self.cursor.execute(f"SELECT dictionary_word_id, number_of_appearances FROM {table_name}")
        self.connection.commit()
        book_table_response = self.cursor.fetchall()
        book_table_appearances = {}
        # Get the number of appearances in the book table
        for i in book_table_response:
            book_table_appearances[i[0]] = i[1]
        # pprint.pprint(book_table_appearances)

        # Add all words to the book_table
        for word, appearances in word_appearances.items():

            dictionary_word_id = word_id_dict[word]

            # If the word is already in the dictionary, add the number of appearances to it
            if dictionary_word_id in book_table_appearances:
                file_appearances = word_appearances[word]
                db_appearances = book_table_appearances[dictionary_word_id]
                print("File app:", file_appearances)
                print("Book app:", book_table_appearances[dictionary_word_id])
                sum = file_appearances + db_appearances
                self.cursor.execute(f"UPDATE {table_name} SET number_of_appearances = ? "
                                    f"WHERE dictionary_word_id = ?", [sum, dictionary_word_id])
                self.connection.commit()

            else:
                self.cursor.execute(f"INSERT INTO {table_name} (dictionary_word_id, number_of_appearances) "
                                    f"VALUES (?,?)", [dictionary_word_id, word_appearances[word]])
                self.connection.commit()

        self.logger.info("Done adding file to database")

    def add_book(self, bookpath: str, book_name: str):
        """
        Adds a file to the autoanki database. This involves the following steps:\n
        1 - Add book to the book_list table
        2 - Add all the files in "bookpath" to the definitions table and book table
        3 - Add book to book_list property

        If given a directory, it will recursively search for all files in the directory and add them.

        if not already there, adding the
        :param bookpath: The filepath to the book. This is file, or a directory of files
        :param book_name: The name of the book. This will show up in the Anki deck
        :return: None
        """
        # TODO Make this work for multiple files. Right now only works for one filepath
        self.logger.info("Adding book...")
        # Gets a 'table name' clean version of the book name
        book_tablename = self.convert_to_tablename(book_name)

        # Add the name of the book to the book_list table
        success = self._create_book_table(book_name, book_tablename)
        if not success:
            self.logger.error("Failed to create book table")
            return

        # Add all the words in the book to the 'definitions' table
        self.add_file_to_database(bookpath, book_tablename)

        # # Add all the words in the bookpath to a new table with the name of the book.
        # success = self.add_book_table_to_db(bookpath, table_book_name)
        # if success is False:
        #     return

        self.logger.info("Done adding book.")
        return True

    def print_info(self):
        """
        Print basic information about the database
        :return:
        """
        self.cursor.execute("SELECT word FROM dictionary")
        all_rows = self.cursor.fetchall()
        self.cursor.execute("SELECT word FROM dictionary WHERE definition IS NULL")
        unfinished_rows = self.cursor.fetchall()
        format_string_int = "{:<30} | {:>8}"
        format_string_dec = "{:<30} | {:>8.2}"
        print("------------------------")
        print(self.database_path)
        print("------------------------")
        print(format_string_int.format("Number of books:", len(self.books)))
        print(format_string_dec.format("Database size (MB):", Path(self.database_path).stat().st_size / (1024 * 1024)))
        print("Dictionary Table:")
        print(format_string_int.format("Number of rows:", len(all_rows)))
        print(format_string_int.format("Number of unfinished rows:", len(unfinished_rows)))

    def update_definition(self, params: list):
        f"""
        Complete a definition for one word in the dictionary table\n
        traditional_script = params[0]\n
        word_type = params[1]\n
        pinyin = params[2]\n
        pinyin_numbers = params[3]\n
        sub_components = params[4]\n
        hsk_level = params[5]\n
        top_level = params[6]\n
        definition = params[7]\n
        word = params[8]
        :param params: A list of params for the database:
        :return:
        """

        self.cursor.execute("UPDATE dictionary "
                            "SET word_traditional = ?, "
                            "word_type = ?,"
                            "pinyin = ?, "
                            "pinyin_numbers = ?,"
                            "sub_components = ?,"
                            "hsk_level = ?,"
                            "top_level = ?,"
                            "definition = ?"
                            "WHERE word = ?",
                            params)
        self.connection.commit()

    def get_all_completed_definitions(self):

        self.cursor.execute("SELECT * FROM dictionary WHERE definition IS NOT NULL")
        raw_words = self.cursor.fetchall()

        # pprint.pp(raw_words)
        words = []
        for row in raw_words:
            word = {
                "word_id": row[0],
                "word": row[1],
                "word_traditional": row[2],
                "word_type": row[3],
                "pinyin": row[4],
                "pinyin_numbers": row[5],
                "number_of_strokes": row[6],
                "sub_components": row[7],
                "frequency": row[8],
                "hsk_level": row[9],
                "top_level": row[10],
                "audio_path": row[11],
                "image_path": row[12],
                "definition": row[13]
            }
            # print(word["word"])
            # pprint.pp(word)
            # words.append(word)
            words.append(word)
        # pprint.pp(words)
        return words

    def unfinished_definitions(self) -> int:
        self.cursor.execute("SELECT word FROM dictionary WHERE definition IS NULL")
        unfinished_rows = self.cursor.fetchall()
        return len(unfinished_rows)

    @property
    def books(self):
        connection = sqlite3.connect(self.database_path)
        cursor = connection.cursor()
        cursor.execute("SELECT book_name FROM book_list")
        return_array = []
        for table in cursor.fetchall():
            return_array.append(table[0])
        return return_array

    @books.setter
    def books(self, value):
        self._books= value
